﻿uses Graph3D, System.Windows.Media;


const
  // размеры клеток:
  CELL_SIZE = 1.0;
  // размеры куба:
  CUBE_SIZE = 0.9;  
  // ширина поля в клетках:
  COLS = 21;
  // высота поля в клетках:
  ROWS = 21;
  // состояние клетки:
  EMPTY = 0; // - пустая 
  FREE = 1;  // - свободная клетка (проход, коридор)      
  PATH = 2;  // - клетка, по которой проходит путь
  NOPATH = 4;// - клетка, по которой нет пути
  WALL = 15; // - стена
  COIN = 2;  // - монета
  КЛЮЧ = 3;   // - ключ
  
  // направления из клетки:  
  UP = 0;   // - вверх
  RIGHT = 1;// - вправо
  DOWN = 2; // - вниз
  LEFT = 3; // - влево
  
  // координаты стартовой клетки:
  START_COL = 1;
  START_ROW = (ROWS - 1) - 1;
  // координаты финишной клетки:
  END_COL = (COLS - 1) - 1;
  END_ROW = 1;

var  
  // Игрок:                       
  player : BoxT;
  // стена:
  block : CubeT;   
  // грунт:
  cell : RectangleT;   
  
  // материалы для грунта и стен:
  mat := ImageMaterial('Cells1.png', 0.25, 0.25);
  matCell := ImageMaterial('Cells2.png', 0.25, 0.25); 
  
  // Лабиринт:
  labyrinth: array [,] of integer := new integer[COLS, ROWS]; 
  cells: array [,] of RectangleT := new RectangleT[COLS, ROWS];   
  blocks: array [,] of CubeT := new CubeT[COLS, ROWS]; 

  // направления из текущей клетки:
  direction: array of integer := new integer[5];  
  
  // цвет пути Игрока:
  PLAYER_PATH_COLOR := GColor.FromArgb(111, 255, 244, 0);
  // стек для хранения координат клеток:
  _stack: Stack<integer> := new Stack<integer>();
  // клетки пути Игрока:
  playerPath: HashSet<(integer, integer)> := new HashSet<(integer, integer)>();
  // координаты последней клетки пути:   
  lastCell: (integer, integer);
  
  // флажок окончания игры:
  flgGameOver: boolean := true;
  // поздравительная табличка:
  imageWin : RectangleT;
  
  // длина пути:
  lenPath := 0;
  // длина тупиков:
  lenTupik := 0;
  // цвет клеток пути:
  PATH_COLOR := GColor.FromArgb(55, 0, 255, 0);
  // цвет тупиковых клеток:
  NOPATH_COLOR := GColor.FromArgb(55, 255, 0, 0); 


  // ПЕЧАТАЕМ ИНФОРМАЦИЮ НА ЭКРАНЕ
  procedure PrintInfo();
  begin  
    var str := 'Длина пути Игрока = ' + playerPath.Count + 
               ' Длина пути: ' + lenPath + 
               ' Длина тупиков: ' + lenTupik;
    View3D.Title := str;
  end;
  
  // ВОЗВРАЩАЕТ X-КООРДИНАТУ КЛЕТКИ
  function GetCellX(col : integer): integer;
  begin
     Result := integer(col * CELL_SIZE) + 9;
  end;
  // ВОЗВРАЩАЕТ Y-КООРДИНАТУ КЛЕТКИ
  function GetCellY(row : integer): integer;
  begin
     Result := integer(row * CELL_SIZE) + 9;
  end;
    
  // СОЗДАЁМ НОВЫЙ ЛАБИРИНТ
  procedure MakeLabyrinth();
  begin
     // всего пустых клеток в Лабиринте:
     var emptyCells := 0;

     // расставляем стены во всех клетках:
     for var r := 0 to ROWS-1 do
        for var c := 0 to COLS-1 do
        begin
          labyrinth[c, r] := WALL;
          
          // пустые клетки:
          if (c * r mod 2 = 1) then
          begin
             labyrinth[c, r] := EMPTY;
             emptyCells += 1; 
          end;
          
          // создаём клон префаба и добавляем в список cells
          // вычисляем координаты клетки:
          var x := c;
          var y := r;        
          var z := -0.01;
  
          var ce := cell.Clone();
          ce.Position := P3D(x, y, z);
          cells[c, r] := ce;           
        end;

        // алгоритм построения лабиринта DFS
        // ----------------------------------

        // старт - в верхнем левом углу:
        var col := 1;
        var row := (ROWS - 1) - 1;
        cells[col, row].Color := Colors.Red;
        
        // старт - в случайной клетке:
        repeat   
           col := Random(COLS);
           row := Random(ROWS);
        until (col * row mod 2 = 1);
        labyrinth[col, row] := FREE;
        
        // одна клетка Лабиринта найдена:
        var nDone := 1;
        
        // очищаем стек:
        _stack.Clear();
        // добавляем в него первую клетку:
        _stack.Push(col * 1000 + row);
        
        // пробиваем проходы между пустыми клетками:
        while (nDone < emptyCells) do
        begin
            // считаем стены возле текущей клетки, 
            // после которых есть ещё не посещённая клетка -->
            var colrow := _stack.Peek();
            // число стен:
            var nWall := 0;
            // снизу:
            col := colrow div 1000;
            row := colrow mod 1000 - 1;
            if (row > 1) and (labyrinth[col, row] = WALL) and
               (labyrinth[col, (row - 1)] = EMPTY) then
            begin
                nWall += 1;
                direction[nWall] := DOWN;
            end;
            // справа:
            col := colrow div 1000 + 1;
            row := colrow mod 1000;
            if (col < COLS - 2) and
               (labyrinth[col, row] = WALL) and
               (labyrinth[col + 1, row] = EMPTY) then
            begin
                nWall += 1;
                direction[nWall] := RIGHT;
            end;
            // сверху: 
            col := colrow div 1000;
            row := colrow mod 1000 + 1;
            if (row < ROWS - 2) and
               (labyrinth[col, row] = WALL) and
               (labyrinth[col, (row + 1)] = EMPTY) then
            begin
                nWall += 1;
                direction[nWall] := UP;
            end;
            // слева:   
            col := colrow div 1000 - 1;
            row := colrow mod 1000;
            if (col > 1) and
               (labyrinth[col, row] = WALL) and
               (labyrinth[col - 1, row] = EMPTY) then
            begin
                nWall += 1;
                direction[nWall] := LEFT;
            end;

            // есть стены:
            if (nWall > 0) then
            begin
                nDone += 1;
                // выбираем случайное направление из возможных:
                var rndDir := Random(nWall) + 1;
                var dir := direction[rndDir];

                case (dir) of
                    // идём вниз:
                    DOWN:
                    begin
                        col := colrow div 1000;
                        row := colrow mod 1000 - 1;
                        labyrinth[col, row] := FREE;
                        labyrinth[col, (row - 1)] := FREE;
                        _stack.Push(col * 1000 + row - 1);
                    end;
                    // идём направо:
                    RIGHT:
                    begin
                        col := colrow div 1000 + 1;
                        row := colrow mod 1000;
                        labyrinth[col, row] := FREE;
                        labyrinth[col + 1, row] := FREE;
                        _stack.Push((col + 1) * 1000 + row);
                    end;
                    // идём вверх:
                    UP:
                    begin
                        //print("SOUTH");
                        col := colrow div 1000;
                        row := colrow mod 1000 + 1;
                        labyrinth[col, row] := FREE;
                        labyrinth[col, row + 1] := FREE;
                        _stack.Push(col * 1000 + row + 1);
                    end;
                    // идём влево:
                    LEFT:
                    begin
                        //print("WEST");
                        col := colrow div 1000 - 1;
                        row := colrow mod 1000;
                        labyrinth[col, row] := FREE;
                        labyrinth[col - 1, row] := FREE;
                        _stack.Push((col - 1) * 1000 + row);
                    end;
                end;
            end
            // нет стен:
            else
                _stack.Pop();
            
        // Лабиринт построен:
        if (nDone = emptyCells) then
        begin
            for var r := 0 to ROWS-1 do
                for var c := 0 to COLS-1 do
                begin
                    if (labyrinth[c, r] <> WALL) then
                        continue;

                    // создаём клон префаба и добавляем в список Cells:
                    // вычисляем координаты клетки:
                    var x := c;
                    var y := r;        
                    var z := 0.5;
                    var bl := block.Clone();         
                    bl.Position := P3D(x, y, z);
                    blocks[c, r] :=  bl;  
                end;
            //Println('Лабиринт готов!!');
        end;          
      end;
        
      // финишная клетка зелёная:
      cells[END_COL, END_ROW].Color := Colors.Green;
      
      // прячем заготовки оснований и стен:
      block.Visible := false;
      cell.Visible := false;        
    end;
    
    
  // ИЩЕМ ПУТЬ В Лабиринте
  procedure FindPath(); 
  begin
     // очищаем стек:
     _stack.Clear();
     // сбрасываем флажок:
     var flg := false;
     // путей пока нет:
     lenPath := 0;
     lenTupik := 0;
     
     // старт в верхнем левом углу:
     var col := START_COL;
     var row := START_ROW;
     // клетка пройдена:
     labyrinth[col, row] := PATH;
     // запоминаем координаты входа на стеке:
     _stack.Push(1000 * col + row);
     lenPath += 1;
     
     while (true) do
     begin
        //PrintInfo();
        // координаты текущей клетки пути:
        var colrow := _stack.Peek();
        col := colrow div 1000;
        row := colrow mod 1000;

        if (col = END_COL) and (row = END_ROW) then
        begin
           // финишная клетка зелёная:
           cells[END_COL, END_ROW].Color := Colors.Green;
           break;
        end;

        flg := false;
        // ВВВЕРХ:
        if (row < ROWS - 1) and (labyrinth[col, row + 1] = FREE) then
        begin
           //print("ВВВЕРХ");
           labyrinth[col, row + 1] := PATH;
           cells[col, row + 1].Color := PATH_COLOR;
           _stack.Push(col * 1000 + row + 1);
           flg := true;
        end
        // ВПРАВО:
        else if (col < COLS - 1) and (labyrinth[col + 1, row] = FREE) then
        begin
           //print("ВПРАВО");
           labyrinth[col + 1, row] := PATH;
           cells[col + 1, row].Color := PATH_COLOR;
           _stack.Push((col + 1) * 1000 + row);
           flg := true;
        end
        // ВНИЗ: 
        else if (row > 1) and (labyrinth[col, row - 1] = FREE) then
        begin
           //print("ВНИЗ");
           labyrinth[col, row - 1] := PATH;
           cells[col, row - 1].Color := PATH_COLOR;
           _stack.Push(col * 1000 + row - 1);
           flg := true;
         end
         // ВЛЕВО:   
         else if (col > 1) and (labyrinth[col - 1, row] = FREE) then
         begin
            //print("ВЛЕВО");
            labyrinth[col - 1, row] := PATH;
            cells[col - 1, row].Color := PATH_COLOR;
            _stack.Push((col - 1) * 1000 + row);
            flg := true;
         end;

         // нашли продолжение пути:
         if (flg) then
         begin
            // клетка пройдена:
            lenPath += 1;
         end
            // нет проходов:
            else
            begin
               // возвращаемся на 1 клетку назад:
               lenPath -= 1;
               lenTupik += 1;
               colrow := _stack.Pop();
               col := colrow div 1000;
               row := colrow mod 1000;
               labyrinth[col, row] := NOPATH;
               cells[col, row].Color := NOPATH_COLOR;
            end;
        end;
     PrintInfo();
  end;
        
        
    // НАЧИНАЕМ НОВУЮ ИГРУ
    procedure NewGame();
    begin
       // прячем табличку:
       imageWin.Visible := false;

       // уничтожаем клетки из предыдущей игры:
       if (cells <> nil) and (cells.Length > 0) then
          begin
            for var row := 0 to ROWS - 1 do
              for var col := 0 to COLS - 1 do  
              begin
                var ce := cells[col, row];
                if (ce <> nil) then
                   ce.Destroy;
                end;
            end;
          
        if (blocks <> nil) and (blocks.Length > 0) then
          begin
            for var row := 0 to ROWS - 1 do
              for var col := 0 to COLS - 1 do  
              begin
                var bl := blocks[col, row]; 
                if (bl <> nil) then
                   bl.Destroy;
                end;
            end;
            
        // создаём Лабиринт:
        MakeLabyrinth();
        
        // очищаем список координат клеток Игрока:
        playerPath.Clear();
        // и добавляем в него стартовую клетку:
        playerPath.Add((START_COL, START_ROW));      
        lastCell := (START_COL, START_ROW);
        
        // устанавливаем Игрока в исходную позицию:
        player.Position := P3D(1, (ROWS - 1) - 1, 1); 
        // печатаем информацию:
        PrintInfo();
        
        // ищем путь в Лабиринте:
        FindPath();  
        
        // игра началась:
        flgGameOver := false;
  end;
    
    
  // СОЗДАЁМ ИГРОКА
  procedure CreatePlayer();
  begin
    player := Box(1, (ROWS - 1) - 1, 1, 
                  CUBE_SIZE, CUBE_SIZE, CUBE_SIZE * 2, 
                  ImageMaterial('wood1.jpg'));
  end;
  
  
  // ЗАКРАШИВАЕМ ЗАДАННУЮ КЛЕТКУ
  procedure ColorCell(col, row: integer; clr: GColor);
  begin
    if (cells[col, row].Color <> GColor.FromArgb(0,0,0,0)) then
      exit;
     cells[col, row].Color := clr;     
  end;
  
  // ПРОВЕРЯЕМ, НЕ ЗАКОНЧЕНА ЛИ ИГРА
  function IsGameOver(col, row: integer) : boolean;
  begin
    var res := (col = END_COL) and (row = END_ROW);
    if (res) then
      begin
        flgGameOver := true;
        Println('ПОЗДРАВЛЯЮ!');
        // показываем надпись:
        imageWin.Visible := true;
      end;
    Result := res;
  end;
  
  // ХОД ИГРОКА
  procedure PlayerMove(dir : integer);
  begin
    // игра закончена:
    if (flgGameOver) then
      exit;
      
     // координаты текущей клетки пути:
     var colrow := lastCell;
     var col := colrow.Item1;
     var row := colrow.Item2;     

     // ВЛЕВО:   
     case dir of
       LEFT:
       begin
          if (labyrinth[col - 1, row] <> WALL) then
          begin
             //Println('ВЛЕВО');
             labyrinth[col - 1, row] := PATH;
             ColorCell(col - 1, row, PLAYER_PATH_COLOR);
             lastCell := (col - 1, row);
             playerPath.Add((col - 1, row));
             player.AnimMoveByX(-CELL_SIZE, 0.25).Begin;
          end;
       end;
       RIGHT:
       begin
          if (labyrinth[col + 1, row] <> WALL) then
          begin
             //Println('ВПРАВО');
             labyrinth[col + 1, row] := PATH;
             ColorCell(col + 1, row, PLAYER_PATH_COLOR);
             playerPath.Add((col + 1, row));    
             lastCell := (col + 1, row);
             player.AnimMoveByX(CELL_SIZE, 0.25).Begin;
          end
       end;
       DOWN:
       begin
          if (labyrinth[col, row - 1] <> WALL)  then
          begin
             //Println('ВНИЗ');
             labyrinth[col, row - 1] := PATH;
             ColorCell(col, row - 1, PLAYER_PATH_COLOR);
             playerPath.Add((col, row - 1));        
             lastCell := (col, row - 1);
             player.AnimMoveByY(-CELL_SIZE, 0.25).Begin;
          end
       end;  
       UP:
       begin
          if (labyrinth[col, row + 1] <> WALL) then
          begin
             //Println('ВВЕРХ');
             labyrinth[col, row + 1] := PATH;
             ColorCell(col, row + 1, PLAYER_PATH_COLOR);
             playerPath.Add((col, row + 1));     
             lastCell := (col, row + 1);
             player.AnimMoveByY(CELL_SIZE, 0.25).Begin;
          end; 
       end;       
    end;
    PrintInfo();
    // проверка:
    IsGameOver(lastCell.Item1, lastCell.Item2);    
  end;
  
  
begin
  Window.Title := 'Как выбраться из Лабиринта';
  Window.SetSize(740, 740);
  Window.Maximize;
  // цвет фона:
  //View3D.BackgroundColor := Colors.Black;

  // настраиваем камеру:
  var z := 51.0;
  Camera.Position := P3D(9, 10, z);  
  Camera.LookDirection := V3D(0, 0, -z);
  Camera.UpDirection := V3D(0, 1, 0);

  View3D.ShowCameraInfo := true;
  View3D.ShowGridLines := false;
  View3D.ShowCoordinateSystem := false;
  View3D.ShowViewCube := false; 
  GridLines.Width := COLS;
  GridLines.Length := ROWS; 
  
  // включаем свет:
  Lights.AddDirectionalLight(Colors.White,V3D(-500,0,-500));
  
  // создаём Игрока:
  CreatePlayer(); 
  
  // клавиши управления Игроком:
  OnKeyDown := k -> 
    case k of
      Key.Left:  PlayerMove(LEFT);  
      Key.Right: PlayerMove(RIGHT);
      Key.Up:    PlayerMove(UP);
      Key.Down:  PlayerMove(DOWN);
    end;
    
    // поздравительная табличка:
    var matr := ImageMaterial('imgVictory.jpg');
    imageWin := Rectangle3D(9.25, 9, 1.5, 2.25, 10, matr);
    imageWin.Rotate(V3D(0,0,1), -90); 
    
    // следующая игра начинается
    // после щелчка по поздравительной табличке:
    OnMouseDown += procedure (x,y,mb) -> begin
      if mb<>1 then exit;
      var v := FindNearestObject(x,y);
      if v = imageWin then
        NewGame();          
    end;       
    
    // строительные блоки:
    block := Cube(0, 0, CUBE_SIZE, CUBE_SIZE, mat);            
    cell := Rectangle3D(0, 0, 0, CUBE_SIZE, CUBE_SIZE, matCell);

    // начинаем новую игру:
    NewGame();
end.